import click

from tensorkube.constants import CliColors
from tensorkube.services.aws_service import check_and_install_aws_cli
from tensorkube.services.eksctl_service import check_and_install_eksctl
from tensorkube.services.istio import check_and_install_istioctl
from tensorkube.services.kubectl_utils import check_and_install_kubectl, check_and_install_helm


def check_and_install_cli_tools():
    try:
        # check and install aws cli
        check_and_install_aws_cli()
        # check and install eksctl
        check_and_install_eksctl()
        # check if kubectl is present and if not install
        check_and_install_kubectl()
        # check and install istioctl
        check_and_install_istioctl()
        # check and install helm
        check_and_install_helm()
        return True
    except Exception as e:
        text = f"Error {e} while installing CLI tools."
        click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
        return False



import time
from datetime import datetime, timezone
import sys
import importlib.metadata

import click

from tensorkube.constants import REGION, Events, DEFAULT_NAMESPACE, get_cluster_name, CliColors, \
    get_cfn_base_stack_name, TeardownType
from tensorkube.constants import get_base_login_url
from tensorkube.helpers import sanitise_assumed_role_arn, track_event, sanitise_name
from tensorkube.migration_service.migration_manager.migration_service import migrate_tensorkube
from tensorkube.services.aws_service import get_aws_account_id, get_aws_user_arn, \
    get_aws_user_principal_arn, get_session_region, check_and_install_aws_cli, get_existing_bucket_name
from tensorkube.services.cloudformation_service import get_validation_records_from_events, \
    get_validation_records_with_wait, get_stack_status_from_stack_name, determine_teardown_type, CfnStackStatus
from tensorkube.services.configure_service import legacy_configure, cfn_configure
from tensorkube.services.dataset_service import upload_tensorkube_dataset, delete_tensorkube_dataset
from tensorkube.services.deploy import Config, deploy_knative_service, DEFAULT_GPUS, DEFAULT_CPU, DEFAULT_MEMORY, \
    DEFAULT_MIN_SCALE, DEFAULT_MAX_SCALE, DEFAULT_ENV, DEFAULT_SECRET, build_app, create_deployment_details
from tensorkube.services.dynamodb_service import get_all_job_statuses, get_job_statuses
from tensorkube.services.ecr_service import delete_all_tensorkube_ecr_repositories
from tensorkube.services.eks_service import update_eks_kubeconfig, give_eks_cluster_access

from tensorkube.services.environment_service import create_new_environment, list_environments, delete_environment
from tensorkube.services.iam_service import create_iam_user, create_eks_access_policy, attach_policy_to_user, \
    create_access_key, create_user_sqs_access_policy, create_user_s3_access_policy, create_user_dynamo_access_policy, \
    has_configure_permissions
from tensorkube.services.job_queue_service import deploy_job, queue_job, get_job_status, delete_all_job_resources, \
    is_existing_queued_job
from tensorkube.services.k8s_service import create_secret, list_secrets, delete_secret, list_keda_scaled_jobs, \
    add_user_to_k8s
from tensorkube.services.kubectl_utils import check_and_install_helm, check_and_install_kubectl
from tensorkube.services.local_service import check_and_install_cli_tools
from tensorkube.services.metapod_service import list_instances_in_cloud, start_devcontainer, purge_devcontainer, \
    pause_devcontainer, reset_cloud
from tensorkube.services.rich_service import list_tensorkube_deployments, describe_deployment, display_deployment_logs, \
    ssh_into_deployed_service, display_secrets, list_tensorkube_datasets, display_job_logs, \
    delete_tensorkube_job_by_prefix, list_tensorkube_training_jobs, display_queued_jobs, get_tensorkube_training_job, \
    prettify_aws_user_keys, display_specific_queued_job, display_dns_records_table, delete_deployment
from tensorkube.services.teardown_service import legacy_teardown, cfn_teardown
from tensorkube.services.tls_service import configure_certificate, get_certificate_stack_name, attach_domain_name, \
    validate_domain, check_if_a_domain_is_validated
from tensorkube.services.train import axolotl_train, get_job_prefix_from_job_id
from tensorkube.tensorkube_token_server import run_server
from tensorkube.services.nydus import get_nydus_image_url
from tensorkube.services.ecr_service import get_or_create_ecr_repository
from tensorkube.services.version_service import print_tensorkube_version


@click.group()
@click.pass_context
def tensorkube(ctx):
    args = sys.argv
    full_command = ""
    subcommand = ctx.invoked_subcommand
    if subcommand:
        if len(args) > 1:
            for arg in args[1:]:
                if arg.startswith("--"):
                    if subcommand == "dev" and arg == "--cloud":
                        full_command += arg + " "
                    else:
                        break
                elif subcommand == "secret" and (arg == "create" or arg == "delete"):
                        full_command += arg + " "
                        break
                else:
                    full_command += arg + " "

        track_event(Events.COMMAND_RUN.value, {"command": full_command,
                                               "tensorkube_version": importlib.metadata.version("tensorkube")})

    pass


@tensorkube.group()
def list():
    """
    List Tensorkube apps that are currently deployed.
    """
    pass


@tensorkube.group()
def secret():
    """Create, list, and delete secrets."""
    pass


@list.command()
@click.option('--env', default=None, type=str, help='Environment to list the services from.')
@click.option('--all', is_flag=True, help='List all services from all environments.')
@click.option('--old', is_flag=True, help='List older sslip urls for using as hosts if you are on an older version.')
def deployments(env, all, old):
    list_tensorkube_deployments(env_name=env, all=all, old=old)


@list.command()
@click.option('--env', default=DEFAULT_NAMESPACE, type=str, help='Environment to list the secrets from.')
def secrets(env):
    secrets = list_secrets(env)
    if not secrets:
        if env == DEFAULT_NAMESPACE:
            click.echo("No secrets found")
        else:
            click.echo(f"No secrets found in environment " + click.style(env, bold=True))
        return
    display_secrets(secrets, env)


@tensorkube.command()
def install_prerequisites():
    """Install all the prerequisites required to run Tensorfuse on a local machine."""
    check_and_install_cli_tools()


@tensorkube.command()
def upgrade():
    """Migrate your tensorfuse cluster to the latest version of tensorkube.

    This command automatically takes all the necessary actions to upgrade the software that is  managing your Tensorfuse cluster.
    """
    migrate_tensorkube()


@tensorkube.command()
@click.option('--vpc-public-subnets', type=str, help='[Experimental] List of public subnets should be given if you want to create in existing VPC. If not given, a new VPC will be created. You should ignore this flag unless you want to create cluster in existing vpc. Tensorfuse recommends to create a new vpc for the tensorkube cluster instead of deploying in existing vpc. This is a enterprise feature. Reach out to tensorfuse for more information.')
@click.option('--vpc-private-subnets', type=str, help='[Experimental] List of private subnets should be given if you want to create in existing VPC. If not given, a new VPC will be created.You should ignore this flag unless you want to create cluster in existing vpc. Tensorfuse recommends to create a new vpc for the tensorkube cluster instead of deploying in existing vpc. This is a enterprise feature. Reach out to tensorfuse for more information.')
@click.option('--legacy', is_flag=True, help='Use the legacy configuration.')
@click.option('--test', is_flag=True, help='Deploy using test configurations.')
def configure(vpc_public_subnets, vpc_private_subnets, legacy, test):
    """Configure the Tensorkube runtime on your private cloud
    """
    click.echo("Configuring the Tensorfuse runtime for your cloud...")
    start_time = time.time() * 1000
    can_track = track_event(Events.CONFIGURE_START.value, {"start_time": start_time})
    if not can_track:
        click.echo("You need to login to Tensorfuse to configure Tensorkube.")
        click.echo("Please login using the command: " + click.style("tensorkube login", fg='cyan'))
        return


    if vpc_public_subnets and vpc_private_subnets:
        if not legacy:
            info_text = click.style("Configuring Tensorkube with existing VPCs is not supported in this version of configure. Please use the --legacy flag to use the legacy configuration.", fg=CliColors.WARNING.value)
            click.echo(info_text)

            end_time = time.time() * 1000
            track_event(Events.CONFIGURE_END.value,
                        {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "vpc_not_supported"})
            return
    confirmation_txt = click.style("Confirmation: Tensorfuse sets up resources such as a Kubernetes cluster, load balancer, and cluster autoscaler to enable serverless GPUs and features like job queues. The estimated base cost for this setup is approximately ") + click.style("$300 per month", bold=True) + click.style(", billed by AWS. Please select \"Yes\" if you wish to proceed.")
    confirmation = click.confirm(confirmation_txt, default=False)
    if not confirmation:
        end_time = time.time() * 1000
        track_event(Events.CONFIGURE_END.value,
                {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "confirmation_declined"})
        return
    session_region = get_session_region()
    if session_region != REGION:
        end_time = time.time() * 1000
        track_event(Events.CONFIGURE_END.value,
                {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "region_not_supported", "region": session_region})
        click.echo(click.style( f"Error: The current region is {session_region}. Tensorfuse currently supports only {REGION} region. Either switch your region to {REGION} or reach out to ", bold=True,
                        fg=CliColors.ERROR.value) + click.style( "founders@tensorfuse.io", italic=True, bold=True, fg=CliColors.ERROR.value) + click.style(" for support.", bold=True, fg=CliColors.ERROR.value))
        return

    has_permissions = has_configure_permissions()
    if not has_permissions:
        end_time = time.time() * 1000
        click.echo("Insufficient AWS permissions to run the configure command. Please contact your AWS admin.")
        track_event(Events.CONFIGURE_END.value,
                {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "insufficient_permissions"})
        return

    click.echo("All permissions are in place. Starting configure...")
    if legacy:
        try:
            success = legacy_configure(vpc_public_subnets, vpc_private_subnets)
            if not success:
                return

            end_time = time.time() * 1000
            track_event(Events.CONFIGURE_END.value,
                        {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time})
            click.echo("Your tensorfuse cluster is ready and you are good to go.")
        except Exception as e:
            #TODO: add exception handling for ctrl+c
            click.echo(f"Error while configuring Tensorfuse: {e}")
            end_time = time.time() * 1000
            track_event(Events.CONFIGURE_END.value,
                        {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "error"})
            return
    else:
        try:
            check_and_install_aws_cli()
            check_and_install_kubectl()
            check_and_install_helm()

            cfn_configure(test)

            status = get_stack_status_from_stack_name(get_cfn_base_stack_name())
            if status != CfnStackStatus.CREATE_COMPLETE:
                click.echo(f"Error while configuring Tensorfuse. Please check the cloudformation stack [{get_cfn_base_stack_name()}] for more details.")
                end_time = time.time() * 1000
                track_event(Events.CONFIGURE_ERROR.value,
                            {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "error"})
                return

            click.echo("Syncing config files for the tensorkube cluster...")
            click.echo("Updating kubeconfig...")
            update_eks_kubeconfig()
            click.echo("Successfully updated the kubeconfig file.")

            end_time = time.time() * 1000
            track_event(Events.CONFIGURE_END.value,
                        {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time})
            click.echo("Your tensorfuse cluster is ready and you are good to go!")
        except Exception as e:
            click.echo("Error while configuring Tensorfuse. {}".format(e))
            end_time = time.time() * 1000
            track_event(Events.CONFIGURE_END.value,
                        {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "error"})
            return


@tensorkube.command()
def account():
    """Get the AWS account ID."""
    click.echo(get_aws_account_id())


@tensorkube.command()
def cluster():
    """Get the current cluster name."""
    click.echo(get_cluster_name())


# The following commands can tear down all the resources that you have created and configured using the CLI.

# uninstall knative
# uninstall istio
# uninstall karpenter
# delete cluster

@tensorkube.command()
def teardown():
    """
    HIGHLY DESTRUCTIVE ACTION!!!! Don't run this command if you don't intend to delete all the resources that you have created and configured using the CLI.

    <Warning>
    This action will remove everything you configured using tensorkube.
    </Warning>
    :return:
    """
    start_time = time.time() * 1000
    track_event(Events.TEARDOWN_START.value, {"start_time": start_time})

    # ask for the secret phrase before tearing down and inform the user this means that you are destroying all the resources
    confirmation = click.confirm(click.style("Are you sure you want to tear down all resources? This will destroy all the resources that you have created and configured using Tensorfuse.", bold=True, fg=CliColors.ERROR.value), default=False)
    if not confirmation:
        click.echo("Tear down aborted.")
        return
    else:
        # ask for  a secret phrase if given secret phrase is not correct then abort the teardown
        secret_phrase = click.prompt(click.style("After this action you will no longer be able to use Tensorfuse. Please enter `accept` to confirm the teardown. ", bold=True,fg=CliColors.ERROR.value), hide_input=True)
        if secret_phrase == 'accept':
            click.echo("Tear down confirmed.")
        else:
            click.echo("Phrase is incorrect. Tear down aborted.")
            return

    click.echo("Tearing down all resources...")

    # TODO?: add logic to delete any other resources
    # TODO?: delete datasets bucket?
    click.echo("Deleting all ECR repositories...")
    delete_all_tensorkube_ecr_repositories()

    teardown_type = determine_teardown_type()
    if teardown_type == TeardownType.LEGACY:
        click.echo("Identified legacy configure. Starting legacy teardown...")

        legacy_teardown()

        end_time = time.time() * 1000
        track_event(Events.TEARDOWN_END.value,
                    {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time})
        click.echo("Tensorfuse has been successfully disconnected from your cluster.")
    elif teardown_type == TeardownType.CFN:
        click.echo("Identified configure. Starting teardown...")
        train_bucket_name = get_existing_bucket_name("keda", "train")
        if train_bucket_name:
            click.echo(click.style(f"You train bucket, ", fg=CliColors.WARNING.value) +
                       click.style(train_bucket_name, fg=CliColors.WARNING.value, bold=True, italic=True) +
                       click.style(" will not be deleted as it might contain your models.", fg=CliColors.WARNING.value)
                       )
            click.echo(click.style("If you wish to delete the bucket, you must do so manually using the AWS console or using the following command:", fg=CliColors.WARNING.value))
            click.echo(click.style(f"aws s3 rb s3://{train_bucket_name}", italic=True, fg=CliColors.WARNING.value))
            click.echo(click.style("Use the ", fg=CliColors.WARNING.value)+
                       click.style("--force", italic=True, fg=CliColors.WARNING.value)+
                       click.style(" flag to delete any non-versioned objects in the bucket", fg=CliColors.WARNING.value)
                       )
            click.echo(click.style("Note: You will not be able to re-configure tensorkube unless you delete this bucket", bold=True, fg=CliColors.ERROR.value))

        cfn_teardown()

        # TODO: figure out where these are made and if we need to delete these in cfn
        # click.echo("Deleting launch templates...")
        # delete_launch_templates()

        end_time = time.time() * 1000
        track_event(Events.TEARDOWN_END.value,
                    {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time})

        # click.echo("Successfully deleted launch templates.")
        stack_status = get_stack_status_from_stack_name(get_cfn_base_stack_name())
        if stack_status == CfnStackStatus.DELETE_COMPLETE or stack_status is None:
            click.echo("Tensorfuse has been successfully disconnected from your cluster.")
        else:
            click.echo(click.style(f"Error: Teardown incomplete. Please check the cloudformation stack [{get_cfn_base_stack_name()}] for more details.", fg=CliColors.ERROR.value))
            click.echo(click.style("Stack deletion incomplete. There may be dangling resources. Please check your AWS console for more details.", fg=CliColors.ERROR.value, bold=True))
            click.echo(click.style("If you are unable to resolve the issue, please reach out to the tensorfuse team for support", fg=CliColors.ERROR.value))
    else:
        click.echo("Unknown teardown type. Please contact support.")
        end_time = time.time() * 1000
        track_event(Events.TEARDOWN_END.value,
                    {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time, "status": "unknown_teardown_type"})


@tensorkube.command()
@click.option('--gpus', default=DEFAULT_GPUS, help='Number of GPUs needed for the service.')
@click.option('--gpu-type', type=click.Choice(['V100', 'A10G', 'T4', 'L4', 'L40S','A100','H100'], case_sensitive=False),
              help='Type of GPU.')
@click.option('--cpu', type=float, default=DEFAULT_CPU, help='Number of CPU millicores. 1000 = 1 CPU')
@click.option('--memory', type=float, default=DEFAULT_MEMORY, help='Amount of RAM in megabytes.')
@click.option('--min-scale', type=int, default=DEFAULT_MIN_SCALE, help='Minimum number of pods to run.')
@click.option('--max-scale', type=int, default=DEFAULT_MAX_SCALE, help='Maximum number of pods to run.')
@click.option('--env', default=DEFAULT_ENV, type=str, help='Environment to deploy the service to.')
@click.option('--github-actions', is_flag=True, help='Deploying from Github Actions.')
@click.option('--secret', default=DEFAULT_SECRET, type=str, multiple=True, help='Secret to use for the deployment.')
@click.option("--efs", is_flag = True, help="Flag to use EFS for the deployment.")
@click.option('--concurrency', type=int, default=100, help='Number of concurrent requests for a single pod to handle at a time. Pods Scale between min and max scale.')
@click.option('--config-file', type=str,
              help='Path to the config.yaml file. you can use this file instead of cli flags. Keys in config.yaml take precedence than cli flags.')
@click.option('--domain-name', type=str, default=None, help='Domain to use for the deployment.')
@click.option('--port', type=int, default=80, help='Port to expose the service on.')
def deploy(gpus, gpu_type, cpu, memory, min_scale, max_scale, env, github_actions, secret, efs, concurrency, config_file, domain_name, port):
    """
    Deploy your containerized application on Tensorkube. This command requires
    a dockerfile to be present in the current directory.
    """
    click.echo('Preparing configurations for deployment...')
    if env == "default":
        env = DEFAULT_ENV
    try:
        config = Config(config_file)
    except Exception as e:
        click.echo(f"Failed to read config.yaml file: {e}")
        return

    config.set_if_not_exist('gpus', gpus).set_if_not_exist('gpu_type', gpu_type).set_if_not_exist('cpu',
                                                                                                  cpu).set_if_not_exist(
        'memory', memory).set_if_not_exist('min_scale', min_scale).set_if_not_exist('max_scale',
                                                                                    max_scale).set_if_not_exist('env',
                                                                                                                env).set_if_not_exist(
        'github_actions', github_actions).set_if_not_exist('secret', secret).set_if_not_exist('domain', domain_name).set_if_not_exist('efs', efs).set_if_not_exist('concurrency', concurrency).set_if_not_exist('port', port)

    # flags validation
    if config.get('gpus') > 0:
        if not config.get('gpu_type'):
            click.echo(click.style("Error: GPU type is required when GPUs are specified.", fg=CliColors.ERROR.value))
            return
    # TODO add domain validation check to see if the domain is configured properly
    if domain_name is not None:
        click.echo(click.style("Validating domain name...", fg=CliColors.INFO.value))
        is_domain_valid = validate_domain(domain=domain_name)
        is_domain_validated = check_if_a_domain_is_validated(domain=domain_name)
        if not (is_domain_valid and is_domain_validated):
            click.echo(click.style("Error: Invalid domain name.", fg=CliColors.ERROR.value))
            return
    start_time = time.time() * 1000
    can_track = track_event(Events.DEPLOY_START.value, {"start_time": start_time})
    if not can_track:
        click.echo("You need to login to Tensorfuse to run your deployment.")
        click.echo("Please login using the command: " + click.style("tensorkube login", fg='cyan'))
        if config.get('github_actions'):
            raise Exception("Please provide the necessary credentials to deploy using tensorkube")
        return
    if config.get('gpus') not in [0, 1, 4, 8]:
        click.echo('Error: Invalid number of GPUs. Only supported values are 0, 1, 4, and 8.')
        return
    click.echo(click.style("Preparing image details", fg=CliColors.INFO.value))
    depl_secrets, env_namespace, image_tag, cwd, sanitised_project_name = create_deployment_details(secret=config.get('secret'),
                                                                                                  env=config.get('env'))
    if not image_tag:
        return
    enable_efs = False
    if config.get('efs'):
        enable_efs = True
    repo_url = get_or_create_ecr_repository(sanitised_project_name)
    image_url = f"{repo_url}:{image_tag}"
    
    convert_to_nydus_image = True  if config.get('gpus') > 0 and not enable_efs else False
    click.echo(click.style("Starting the build process...", fg=CliColors.INFO.value))
    pod_status = build_app(env=config.get('env'), image_tag=image_tag, sanitised_project_name=sanitised_project_name,
                           upload_to_nfs=enable_efs, image_url=image_url, convert_to_nydus_image=convert_to_nydus_image, secrets=config.get('secret'))
    if config.get('gpus') > 0:
        image_url = get_nydus_image_url(image_url=image_url)
    if enable_efs:
        image_url = image_tag
    deploy_knative_service(env=config.get('env'), gpus=config.get('gpus'),
                           gpu_type=config.get('gpu_type'), cpu=config.get('cpu'), memory=config.get('memory'),
                           min_scale=config.get('min_scale'), max_scale=config.get('max_scale'),concurrency= config.get('concurrency'), pod_status=pod_status,
                           cwd=cwd, sanitised_project_name=sanitised_project_name, image_url=image_url, secrets=depl_secrets,
                           enable_efs=enable_efs, domain=config.get('domain'),github_actions=config.get('github_actions'), readiness = config.get('readiness'), port = config.get('port'))

    end_time = time.time() * 1000
    track_event(Events.DEPLOY_END.value,
                {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time})


# @tensorkube.command()
# def delete_project():
#     click.echo("Deleting the project resources...")
#     # TODO!: add logic to delete the ecr repository, s3 folder, build job, and any other resources
#     click.echo("Successfully deleted the project resources.")


@tensorkube.command()
def get_permissions_command():
    """Get the admin command needed to grant access to the current user.

    <Note>
    This command is only needed if you are not the initial user who created the EKS cluster.
    The user that doesn't have permissions needs to run this command and then send the resulting command
    to the admin user that provisioned the Tensorfuse runtime.
    </Note>
    """
    # TODO: give details of cluster user as well
    click.echo(f"Ask the initial user to run this command to grant you the necessary"
               f" permissions to the {get_cluster_name()} EKS cluster:")

    user_arn = get_aws_user_arn()
    if 'assumed-role' in user_arn:
        final_arn = sanitise_assumed_role_arn(user_arn)
    else:
        final_arn = user_arn

    click.echo("""\
    eksctl create iamidentitymapping \\
    --cluster {} \\
    --region {} \\
    --arn {} \\
    --group system:masters \\
    --username <USERNAME_OF_YOUR_CHOICE>""".format(get_cluster_name(), REGION, final_arn))

    click.echo("Once you have access to the cluster, run the following command to sync the config files:")
    click.echo("tensorkube sync")


@tensorkube.command()
def sync():
    """Sync the Tensorfuse configuration on a new machine.
    """
    click.echo("Syncing config files for the tensorkube cluster...")
    click.echo("Updating kubeconfig...")
    update_eks_kubeconfig()
    click.echo("Successfully updated the kubeconfig file.")


@tensorkube.group()
def deployment():
    """Manage and interact with your deployments."""
    pass


@deployment.command()
@click.option('--env', default=None, type=str, help='Environment to list the services from.')
@click.option('--all', is_flag=True, help='List all services from all environments.')
@click.option('--old', is_flag=True, help='List older sslip urls for using as hosts if you are on an older version.')
def list(env, all, old):
    """List all deployments running and recently stopped along with their status"""
    list_tensorkube_deployments(env_name=env, all=all, old=old)


@deployment.command()
@click.argument('service_name')
@click.option('--env', default=None, type=str, help='Environment to describe the deployment from.')
def describe(service_name, env):
    """List details of a deployment including the status, url and the latest ready revision."""
    describe_deployment(service_name=service_name, env_name=env)

@deployment.command()
@click.argument('service_name')
@click.option('--env', default=None, type=str, help='Environment to describe the deployment from.')
def delete(service_name, env):
    """Delete a deployment.

    <Warning>
    This action is destructive. Once a deployment is deleted it cannot be recovered.
    </Warning>
    """
    env_namespace = env if env else DEFAULT_NAMESPACE
    delete_deployment(service_name=service_name, env_name=env_namespace)


@deployment.command()
@click.argument('service_name')
@click.option('--env', default=None, type=str, help='Environment to list the services from.')
def logs(service_name, env):
    """Stream logs for a specific deployment if it is in the running state.

    Examples:
        Get the logs using the deployment name.

        ```bash
        tensorkube deployment logs demo-gpus-1-v100
        ```
    """
    env_namespace = env if env else DEFAULT_NAMESPACE
    display_deployment_logs(service_name=service_name, namespace=env_namespace)


@deployment.command()
@click.argument('service_name')
@click.option('--env', default=None, type=str, help='Environment to list the services from.')
def ssh(service_name, env):
    """SSH into the instances of a specific deployment.

    If multiple instances are running, this command gives you a list of instances to choose from.

    Examples:
        SSH into the deployment instance using deployment name.

        ```bash
        tensorkube deployment ssh demo-gpus-1-v100
        ```
    """
    click.echo("SSH into the deployment...")
    env_namespace = env if env else DEFAULT_NAMESPACE
    ssh_into_deployed_service(service_name=service_name, namespace=env_namespace)


# all environment specific commands
@tensorkube.group()
def environment():
    """
    Environments are subdivisions within workspaces that enable you to deploy the same application across different namespaces.
    Each environment operates independently, with its own set of Secrets. By default, any lookups performed by an application within
    a specific environment are scoped to entities in that same environment.

    A common use case for environments is to separate development and production workflows. This ensures that new features can be
    developed and tested in a dedicated development environment without risking changes to live production applications. At the same time,
    it allows seamless deployment of updates to the production environment when ready.

    <Note>
    Containers from different environments can run on the same nodes. However, they are isolated from each other.
    Apps in different environments still eat up your quotas. So, make sure to not keep the instances always running in your
    staging environments.
    </Note>
    """
    pass


@environment.command()
@click.option('--env-name', help='Name of the environment to create.')
def create(env_name):
    """Create a new environment.

    This also provisions the surrounding infrastructure needed for environments such as service accounts, IAM roles, registries and buckets.
    """
    click.echo("Creating a new environment...")
    create_new_environment(env_name=env_name)


@environment.command()
@click.option('--env-name', help='Name of the environment to delete.')
def delete(env_name):
    """Delete an environment.

    This command deletes all the deployed jobs and deployments in the environment.

    This also de-provisions all the surrounding infrastructure needed for environments such as service accounts, IAM roles, registries and buckets.

    <Warning>
    Please be careful while deleting an environment. This action is irreversible. All the jobs, queues and deployed applications in the environment will be deleted.
    </Warning>
    """
    click.echo("Deleting the environment...")
    delete_environment(env_name=env_name)


@environment.command()
def list():
    """List all the environments.
    """
    environments = list_environments()
    for env in environments:
        click.echo(env)


@tensorkube.command()
def login():
    """Log into Tensorfuse and obtain the usage token.

    Tensorfuse CLI requires obtaining a usage token from Tensorfuse.
    """
    click.echo("Opening the Tensorfuse dashboard in your default browser...")
    base_url = get_base_login_url()
    click.launch(base_url + "/tensorfuse/login?tensorkube_login=True")
    run_server()
    track_event(Events.LOGGED_IN.value, {})



@secret.command()
@click.option('--env', default=DEFAULT_NAMESPACE, type=str, help='Environment to create the secret in.')
@click.option('--force', is_flag=True, help='Update the secret if one with the same name already exists.')
@click.argument('secret_name', type=str)
@click.argument('key_value_pairs', nargs=-1)
def create(secret_name, key_value_pairs, env, force):
    """Create a new secret in the specified environment. Use `--force` to overwrite an existing one.
    """
    try:
        key_values = dict(pair.split('=',1) for pair in key_value_pairs)
    except Exception as e:
        click.echo(
            click.style("Error: Invalid key-value pairs. Please provide key-value pairs in the format KEY=value.",
                        fg=CliColors.ERROR.value))
        return

    click.echo("Creating secret " + click.style(secret_name, bold=True) + " in environment " + click.style(env,
                                                                                                           bold=True) + " ...")

    create_secret(name=secret_name, data=key_values, namespace=env, force=force)


@secret.command()
@click.option('--env', default=DEFAULT_NAMESPACE, type=str, help='Environment to delete the secret from.')
@click.argument('secret_name', type=str)
def delete(secret_name, env):
    """Delete a secret from the specified environment.

    <Warning>
    This action cannot be undone. Deleting a secret will affect all the deployments that are using that particular secret.
    They might completely fail to initialise. Use this command with caution.
    </Warning>
    """
    msg = "Are you sure you want to delete the secret? This action cannot be undone"
    if click.confirm(click.style(msg, fg=CliColors.WARNING.value), abort=True):
        click.echo("Deleting secret " + click.style(secret_name, bold=True) + " from environment " + click.style(env,
                                                                                                                 bold=True) + " ...")

        delete_secret(name=secret_name, namespace=env)


@secret.command()
@click.option('--env', default=DEFAULT_NAMESPACE, type=str, help='Environment to list the secrets from.')
def list(env):
    """List all the secrets in the specified environment.
    """
    secrets = list_secrets(env)
    if not secrets:
        if env == DEFAULT_NAMESPACE:
            click.echo("No secrets found")
        else:
            click.echo(f"No secrets found in environment " + click.style(env, bold=True))
        return
    display_secrets(secrets, env)


@tensorkube.command()
def test():
    click.echo("Running tests... for someone")
    


@tensorkube.command()
@click.option('--cloud', default='aws', help='Cloud provider to use.')
def reset(cloud):
    reset_cloud(cloud)


@tensorkube.group()
@click.option('--cloud', default='aws', help='Cloud provider to use.')
@click.pass_context
def dev(ctx, cloud):
    """Manage hot reloading development containers for the current user.
    """
    ctx.ensure_object(dict)
    ctx.obj['cloud'] = cloud
    time_now_utc = datetime.now(timezone.utc).isoformat()
    subcommand = ctx.invoked_subcommand
    if subcommand == "start":
        event = Events.DEVCONTAINER_START
    elif subcommand == "stop":
        event = Events.DEVCONTAINER_STOP
    elif subcommand == "delete":
        event = Events.DEVCONTAINER_DELETE
    elif subcommand == "list":
        event = Events.DEVCONTAINER_LIST
    else:
        event = Events.DEVCONTAINER
    can_track = track_event(event.value, {"start_time": time_now_utc})
    if not can_track:
        click.echo("You need to login to Tensorfuse to run your devcontainers.")
        click.echo("Please login using the command: " + click.style("tensorkube login", fg='cyan'))
        exit(0)


@dev.command()
@click.pass_context
def list(ctx):
    """List all the running / stopped devcontainers from the current user.
    """
    cloud = ctx.obj['cloud']
    list_instances_in_cloud(cloud)


@dev.command()
@click.option('--gpu-type',
              type=click.Choice(['V100', 'A10G', 'T4', 'L4', 'A100', 'L40S', 'H100', 'None'], case_sensitive=False),
              help='Type of GPU.')
@click.option('--port', type=int, default=8080, help='Port to run the devcontainer on.')
@click.pass_context
def start(ctx, gpu_type: str, port: int):
    """Spin up a new devcontainer from the current folder.

    The folder should contain a `Dockerfile`.
    """
    cloud = ctx.obj['cloud']
    if gpu_type == 'None':
        gpu_type = None
    start_devcontainer(cloud, gpu_type, port)


@dev.command()
@click.pass_context
def stop(ctx):
    """Stop a running development container.

    This is only available when the underlying cloud supports stopped instances. An advantage of stopping against deleting
    a devcontainer is that the SSD snapshot is cached and hence, the devcontainer can be started faster.

    However this incurs cost for the SSD snapshot.
    """
    cloud = ctx.obj['cloud']
    pause_devcontainer(cloud)


@dev.command()
@click.pass_context
def delete(ctx):
    """Purge a running / stopped development container.

    This command deletes all the associated resources of the development container including the attached block storage.
    """
    cloud = ctx.obj['cloud']
    purge_devcontainer(cloud)


@tensorkube.group()
def train():
    pass


@train.command()
@click.option('--gpus', default=DEFAULT_GPUS, help='Number of GPUs needed for the service.')
@click.option('--gpu-type', type=click.Choice(['V100', 'A10G', 'T4', 'L4', 'L40S', 'A100', 'H100'], case_sensitive=False),
              help='Type of GPU.')
@click.option('--env', default=DEFAULT_ENV, type=str, help='Environment to deploy the service to.')
@click.option('--job-id', default=None, type=str, help='Unique job id for the training job.')
@click.option('--axolotl', is_flag=True, help='Run the axolotl training job.')
@click.option('--config-path', default=None, type=str, help='Path to the config.yaml file.')
@click.option('--secret', default=DEFAULT_SECRET, type=str, multiple=True, help='Secret to use for the deployment.')
def create(gpus, gpu_type, env, job_id, axolotl, config_path, secret):
    if axolotl:
        axolotl_train(env=env, secrets=secret, gpus=gpus, gpu_type=gpu_type, job_id=job_id, config_path=config_path)
    else:
        click.echo("Invalid command. Please use the --axolotl flag to run the axolotl training job.")
        return


@train.command()
@click.option('--name', required=True, type=str, help='Name of the user to create.')
def create_user(name: str):
    click.echo("Creating a new user...")
    create_iam_user(name)
    policy_arn = create_eks_access_policy(f'{get_cluster_name()}-{name}-eks-access-policy')['Policy']['Arn']
    attach_policy_to_user(policy_arn, name)
    sqs_policy_arn = create_user_sqs_access_policy(f'{get_cluster_name()}-{name}-sqs-access-policy')['Policy']['Arn']
    attach_policy_to_user(sqs_policy_arn, name)
    s3_policy_arn = create_user_s3_access_policy(f'{get_cluster_name()}-{name}-s3-access-policy')['Policy']['Arn']
    attach_policy_to_user(s3_policy_arn, name)
    dynamo_policy_arn = create_user_dynamo_access_policy(f'{get_cluster_name()}-{name}-dynamo-access-policy')['Policy'][
        'Arn']
    attach_policy_to_user(dynamo_policy_arn, name)
    add_user_to_k8s(name)
    access_key, secret_access_key = create_access_key(name)
    prettify_aws_user_keys(access_key, secret_access_key)
    return


@train.command()
@click.option('--job-id', required=True, type=str, help='Unique job id for the training job.')
@click.option('--env', default=DEFAULT_ENV, type=str, help='Environment to deploy the service to.')
def logs(job_id: str, env: str):
    click.echo("Fetching logs for the training job...")
    job_prefix = get_job_prefix_from_job_id(job_id)
    namespace = env if env else DEFAULT_NAMESPACE
    display_job_logs(job_prefix=job_prefix, namespace=namespace)


@train.command()
@click.option('--env', default=DEFAULT_ENV, type=str, help='Environment to delete the job from.')
@click.option('--job-id', required=True, type=str, help='Unique job id for the training job.')
def delete(job_id: str, env: str):
    click.echo("Deleting the training job...")
    job_prefix = get_job_prefix_from_job_id(job_id)
    namespace = env if env else DEFAULT_NAMESPACE
    delete_tensorkube_job_by_prefix(job_prefix=job_prefix, namespace=namespace)


@train.command()
@click.option('--env', default=DEFAULT_ENV, type=str, help='Environment to deploy the service to.')
def list(env: str):
    namespace = env if env else DEFAULT_NAMESPACE
    # ax- for axolotl jobs
    prefix = 'ax-'
    list_tensorkube_training_jobs(namespace=namespace, job_prefix=prefix)


@train.command()
@click.option('--job-id', required=True, type=str, help='Unique job id for the training job.')
@click.option('--env', default=DEFAULT_ENV, type=str, help='Environment to deploy the service to.')
def get(job_id: str, env: str):
    namespace = env if env else DEFAULT_NAMESPACE
    # ax- for axolotl jobs
    job_prefix = get_job_prefix_from_job_id(job_id)
    get_tensorkube_training_job(job_prefix=job_prefix, namespace=namespace)


@tensorkube.group()
def datasets():
    """Manage datasets for Tensorkube
    """
    pass


@datasets.command()
def list():
    """List all the datasets created using Tensorkube cli.
    """
    list_tensorkube_datasets()


@datasets.command()
@click.option("--dataset-id", required=True, type=str, help="ID of the dataset to create.")
@click.option("--path", required=True,
              type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=str, readable=True),
              callback=lambda ctx, param, value: value if value.endswith('.jsonl') else click.BadParameter(
                  'File must be a .jsonl file'), help="Path to the JSONL file.")
def create(dataset_id, path):
    """Create a dataset using Tensorkube. This command creates a cloud bucket and stores your dataset in that bucket.
    """
    click.echo("Creating dataset...")
    upload_tensorkube_dataset(dataset_id=dataset_id, file_path=path)


@datasets.command()
@click.option("--dataset-id", required=True, type=str, help="ID of the dataset to delete.")
def delete(dataset_id: str):
    """Delete a dataset using Tensorkube cli.
    """
    delete_tensorkube_dataset(dataset_id=dataset_id)


@tensorkube.group()
def job():
    """Manage, deploy and queue jobs on Tensorfuse.
    """
    pass


@job.command()
@click.option('--name', required=True, type=str, help='Name of the job to deploy.')
@click.option('--gpus', default=DEFAULT_GPUS, help='Number of GPUs needed for an instance of the job.')
@click.option('--gpu-type', type=click.Choice(['V100', 'A10G', 'T4', 'L4', 'L40S', 'A100', 'H100'], case_sensitive=False),
              help='Type of GPU.')
@click.option('--cpu', type=float, default=DEFAULT_CPU, help='Number of CPU millicores. 1000 = 1 CPU')
@click.option('--memory', type=float, default=DEFAULT_MEMORY, help='Amount of RAM in megabytes.')
@click.option('--secret', default=DEFAULT_SECRET, type=str, multiple=True, help='Secrets that are required by the job.')
@click.option('--max-scale', type=int, default=DEFAULT_MAX_SCALE, help='Maximum number of jobs to run concurrently.')
@click.option('--update', is_flag=True, help='Update the job if it already exists.')
@click.option("--efs", is_flag = True, help="Flag to use EFS for the deployment.")
def deploy(name, gpus, gpu_type, cpu, memory, secret, max_scale, update, efs):
    """Deploy and orchestrate a job on Tenosrfuse.

    This command orchestrates the infrastructure needed for jobs such as queues, buckets and key-value stores.
    It also builds the container image for the job and pushes it to the container registry.
    """
    env = 'keda'
    start_time = time.time() * 1000
    can_track = track_event(Events.JOB_DEPLOY_START.value, {"start_time": start_time})
    if not can_track:
        click.echo("You need to login to Tensorfuse to run your deployment.")
        click.echo("Please login using the command: " + click.style("tensorkube login", fg='cyan'))
        return
    if gpus not in [0, 1, 4, 8]:
        click.echo('Error: Invalid number of GPUs. Only supported values are 0, 1, 4, and 8.')
        return

    # TODO: Check if job already exists and update it if the update flag is set
    does_job_exist, existing_job = is_existing_queued_job(name)
    if does_job_exist and not update:
        click.echo(
            f"Job with name {name} already exists. Please use a different name or use the --update flag to update the job.")
        return

    depl_secrets, env_namespace, image_tag, cwd, sanitised_project_name = create_deployment_details(secret=secret,
                                                                                                    env=env)

    if not cwd and not image_tag:
        return

    sanitised_job_name = sanitise_name(name)
    repo_url = get_or_create_ecr_repository(sanitised_project_name)
    image_url = f"{repo_url}:{image_tag}"
    convert_to_nydus_image = True if gpus > 0 and not efs else False
    pod_status = build_app(env=env, image_tag=image_tag, sanitised_project_name=sanitised_job_name, upload_to_nfs=efs,image_url=image_url, convert_to_nydus_image=convert_to_nydus_image, secrets=secret)
    if gpus > 0:
        image_url = get_nydus_image_url(image_url=image_url)
    if efs:
        image_url = image_tag
    if pod_status == 'Succeeded' or pod_status == 'Completed':
        deploy_job(env=env, job_name=name, gpus=gpus, gpu_type=gpu_type, cpu=cpu, memory=memory, max_scale=max_scale,
                   sanitised_project_name=sanitised_job_name, image_tag=image_url, secrets=depl_secrets, cwd=cwd, update=update, enable_efs=efs)
    else:
        click.echo("Failed to build the Docker image. Please check the logs for more details. Pod status: {}".format(
            pod_status))

    end_time = time.time() * 1000
    track_event(Events.JOB_DEPLOY_END.value,
                {"start_time": start_time, "end_time": end_time, "duration": end_time - start_time})


@job.command()
@click.option("--job-name", required=True, type=str, help="Name of the job to delete.")
def delete(job_name: str):
    """Delete a job from Tensorfuse.

    This command also de-provisions a job's infrastructure such as queues, buckets and key-value stores.
    """
    delete_all_job_resources(job_name)
    click.echo("Job deleted successfully.")


@job.command()
@click.option("--job-name", required=True, type=str, help="Name of the job to delete.")
@click.option("--job-id", required=True, type=str, help="Unique id for the job instance.")
@click.option("--payload", required=True, type=str, help="Payload of the job to delete.")
def queue(job_name: str, job_id: str, payload: str):
    """Queue an instance of a deployed job on Tensorfuse.

    The payload is a JSON string that is passed to the job instance.
    """
    queue_job(job_name, job_id, payload)
    click.echo("Job queued successfully.")


@job.command()
@click.option("--job-name", required=True, type=str, help="Name of the job to deploy.")
@click.option("--job-id", required=True, type=str, help="Unique id of the job instance.")
def get(job_name: str, job_id: str):
    """Get the status of a job instance.
    """
    status = get_job_status(job_name, job_id)
    print(status)


@job.command()
@click.option("--job-name", default=None, type=str, help="Name of the job to deploy.")
def list(job_name):
    """List all jobs or a specific job on Tensorfuse.

    1. If no job name is provided, all jobs are listed.
    2. If a job name is provided, the status of that job is displayed.
    """
    if not job_name:
        # Display for all jobs
        scaled_jobs = list_keda_scaled_jobs()
        queued_job_statuses = get_all_job_statuses()
        display_queued_jobs(scaled_jobs, queued_job_statuses)
    else:
        # Display for a specific job
        does_job_exist, existing_job = is_existing_queued_job(job_name)
        if not does_job_exist:
            click.echo(f"No job found with name: {job_name}")
            return
        else:
            dynamodb_job_statuses = get_job_statuses(job_name)
            display_specific_queued_job(existing_job, dynamodb_job_statuses)


@job.command()
@click.option("--job-name", required=True, type=str, help="Name of the job whose logs to display.")
def logs(job_name):
    """Stream the logs for a specific job instance.

    This command works only if the job is currently in progress.
    """
    display_job_logs(job_prefix=job_name, namespace='keda', container_name="sqs-job-1")


@tensorkube.group()
def domain():
    """Manage custom domain names and TLS for your Tensorkube cluster.
    """
    pass


@domain.command()
@click.option('--domain-name', required=True, help='Domain name for wildcard certificate')
def configure(domain_name):
    """Configure a wildcard subdomain for your Tensorkube platform.

    This command creates an ACM certificate for your wildcard domain and sets up DNS validation.

    Example:
        To assign your tensorfuse deployments a domain of the form `tfservice.sub.example.org` run the following command
        ```bash
        tensorkube domain configure --domain-name docdemo.tensorfuse.io
        ```
        This will configure a wildcard certificate for all the domains of the form `*.docdemo.tensorfuse.io`

    """
    deployed = configure_certificate(domain_name)
    if not deployed:
        return
    stack_name = get_certificate_stack_name(domain_name)
    validation_records = get_validation_records_with_wait(stack_name=stack_name)
    if validation_records is None:
        click.echo(click.style(f"Failed to retrieve validation records for domain: {domain_name}. "
                               f"Please configure the record using `tensorkube domain configure` or try again (race condition).",
                               fg='red'))
        return
    else:
        display_dns_records_table(domain_name, validation_records)


@domain.command()
@click.option('--domain-name', required=True, help='Domain name for wildcard certificate')
def attach(domain_name):
    """Associate the configured domain to your tensorfuse runtime post DNS validation.
    """
    attach_domain_name(domain_name)  # in the end configure 443 port for the ingress gateway


@domain.command()
@click.option('--domain-name', required=True, help='Domain name for wildcard certificate')
def get_validation_records(domain_name):
    """Get the DNS validation records and DNS routing records for the domain.
    """
    stack_name = get_certificate_stack_name(domain_name)
    validation_records = get_validation_records_from_events(stack_name=stack_name)
    if validation_records is None:
        click.echo(click.style(f"Failed to retrieve validation records for domain: {domain_name}. "
                               f"Please configure the record using `tensorkube domain configure`", fg='red'))
        return
    else:
        display_dns_records_table(domain_name, validation_records)


@tensorkube.command()
@click.option('--principal-arn', required=True, type=str, help='Principal ARN to give access to the cluster.')
def give_cluster_access(principal_arn: str):
    """Grant access to the EKS cluster to a user or role.
    """
    give_eks_cluster_access(principal_arn)


@tensorkube.command()
def get_principal_arn():
    """Display the principal ARN of the current AWS user. Useful when granting access to the cluster.
    """
    click.echo("Your principal ARN is: " + get_aws_user_principal_arn())


@tensorkube.command()
def version():
    print_tensorkube_version()


from tensorkube.services.job_queue_service import create_cloud_resources_for_queued_job_support
import click

def apply():
    try:
        create_cloud_resources_for_queued_job_support()
        click.echo("Successfully Created Cloud Resources for Queued Job Support")
    except Exception as e:
        raise e


import boto3
from tensorkube.constants import REGION, get_cluster_name


def get_dynamodb_table_name():
    return f"{get_cluster_name()}-job-status"

def get_dynamodb_access_policy_name():
    return f"{get_cluster_name()}-dynamo-access-policy"

def get_user_dynamodb_access_policy_name(user: str):
    return f"{get_cluster_name()}-{user}-dynamodb-access-policy"


def get_all_job_statuses(region: str = REGION, limit: int = 3):
    dynamodb = boto3.resource('dynamodb', region_name=region)
    table_name = get_dynamodb_table_name()
    table = dynamodb.Table(table_name)

    all_jobs = {}
    last_evaluated_key = None

    try:
        while True:
            if last_evaluated_key:
                response = table.scan(ExclusiveStartKey=last_evaluated_key)
            else:
                response = table.scan()

            # Extract job_name, job_id, and status for each item
            for item in response.get('Items', []):
                job_name = item.get('job_name')
                if job_name not in all_jobs:
                    all_jobs[job_name] = []

                all_jobs[job_name].append({
                    'job_name': job_name,
                    'job_id': item.get('job_id'),
                    'status': item.get('status'),
                })

            last_evaluated_key = response.get('LastEvaluatedKey')

            if not last_evaluated_key:
                break

        latest_jobs = {}
        for job_name, jobs in all_jobs.items():
            latest_jobs[job_name] = sorted(jobs, key=lambda x: x['job_id'], reverse=True)[:limit]
            if len(all_jobs[job_name]) > limit:
                latest_jobs[job_name].append({
                    'job_name': job_name,
                    'job_id': '...',
                    'status': '...',
                })

        return latest_jobs

    except Exception as e:
        print(f"Error fetching job statuses: {e}")
        return []


def get_job_statuses(job_name: str):
    dynamodb = boto3.resource('dynamodb', region_name=REGION)
    table_name = get_dynamodb_table_name()
    table = dynamodb.Table(table_name)

    response = table.query(
        KeyConditionExpression='job_name = :job_name',
        ExpressionAttributeValues={
            ':job_name': job_name,
        }
    )

    return response.get('Items', [])



import boto3
from tensorkube.constants import REGION, get_cluster_name


def get_sqs_access_policy_name():
    return f"{get_cluster_name()}-sqs-access-policy"


def create_sqs_queue(queue_name: str, region: str = REGION) -> str:
    sqs = boto3.client('sqs', region_name=region)
    response = sqs.create_queue(
        QueueName=queue_name,
        Attributes={
            'DelaySeconds': '0',
            'MessageRetentionPeriod': '86400',  # Retain messages for 1 day
            'VisibilityTimeout': '43200' # Make messages invisible for 12 hrs
        }
    )
    queue_url = response['QueueUrl']
    return queue_url


def delete_sqs_queue(queue_name: str, region: str = REGION) -> None:
    sqs = boto3.client('sqs', region_name=region)
    queue_url = sqs.get_queue_url(QueueName=queue_name)['QueueUrl']
    sqs.delete_queue(QueueUrl=queue_url)
    print(f"Queue {queue_name} deleted successfully.")

def queue_message(queue_url: str, message_body: str, region: str = REGION) -> None:
    sqs = boto3.client('sqs', region_name=region)
    response = sqs.send_message(
        QueueUrl=queue_url,
        MessageBody=message_body
    )
    print(f"Message sent to queue")



import subprocess
import time
from datetime import datetime
from typing import Optional, List, Dict

import click
import yaml
from kubernetes import config, client
from kubernetes.client import ApiException

from tensorkube.constants import KNATIVE_SERVING_NAMESPACE, CONFIG_FEATURES, TENSORFUSE_NAMESPACES, DEFAULT_NAMESPACE
from tensorkube.services.k8s_service import get_efs_claim_name, get_tensorkube_cluster_context_name
from tensorkube.services.nydus import get_nydus_runtime_class_name
from tensorkube.services.nodepool import get_gpu_nodepool


def get_instance_family_from_gpu_type(gpu_type):
    gpu_type = gpu_type.lower()
    if gpu_type == 'v100':
        return 'p3'
    elif gpu_type == 'a10g':
        return 'g5'
    elif gpu_type == 't4':
        return 'g4dn'
    elif gpu_type == 't4g':
        return 'g5g'
    elif gpu_type == 'l4':
        return 'g6'
    elif gpu_type == 'l40s':
        return 'g6e'
    elif gpu_type == 'a100':
        return 'p4d'
    elif gpu_type == 'h100':
        return 'p5'
    else:
        raise ValueError(f"Unsupported GPU type: {gpu_type}")

def get_gpu_type_from_instance_family(instance_family):
    if instance_family == 'p3':
        return 'v100'
    elif instance_family == 'g5':
        return 'a10g'
    elif instance_family == 'g4dn':
        return 't4'
    elif instance_family == 'g5g':
        return 't4g'
    elif instance_family == 'g6':
        return 'l4'
    elif instance_family == 'g6e':
        return 'l40s'
    elif instance_family == 'p4d':
        return 'a100'
    elif instance_family == 'p5':
        return 'h100'
    else:
        raise ValueError(f"Unsupported instance family: {instance_family}")


def get_supported_gpu_families():
    return ['p3', 'g5', 'g4dn', 'g5g', 'g6', 'g6e']


def apply_anti_affinity_for_gpus(yaml_dict):
    # this function assumes that already there are anti affinity terms
    existing_terms = \
    yaml_dict['spec']['template']['spec']['affinity']['nodeAffinity']['requiredDuringSchedulingIgnoredDuringExecution'][
        'nodeSelectorTerms']
    for term in existing_terms:
        for expression in term.get('matchExpressions', []):
            if expression['key'] == 'karpenter.k8s.aws/instance-family' and expression['operator'] == 'NotIn':
                expression['values'] = list(set(expression['values'] + get_supported_gpu_families()))
    return yaml_dict


def update_or_apply_istio_gateway(gateway_name: str, gateway_namespace: str,
                                  gateway_yaml_dict: Dict):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        raise ValueError("Could not get the cluster context name")

    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)

    # Define Istio Gateway API parameters
    group = "networking.istio.io"
    version = "v1"
    plural = "gateways"

    try:
        # Check if Gateway exists
        existing_gateway = k8s_client.get_namespaced_custom_object(
            group=group,
            version=version,
            namespace=gateway_namespace,
            plural=plural,
            name=gateway_name
        )

        # Update existing Gateway
        resource_version = existing_gateway['metadata']['resourceVersion']
        gateway_yaml_dict['metadata']['resourceVersion'] = resource_version
        click.echo(f'Resource version is {resource_version}')

        k8s_client.patch_namespaced_custom_object(
            group=group,
            version=version,
            namespace=gateway_namespace,
            plural=plural,
            name=gateway_name,
            body=gateway_yaml_dict
        )
        click.echo(f"Updated Gateway {gateway_name}")

    except ApiException as e:
        if e.status == 404:
            # Create new Gateway if it doesn't exist
            k8s_client.create_namespaced_custom_object(
                group=group,
                version=version,
                namespace=gateway_namespace,
                plural=plural,
                body=gateway_yaml_dict
            )
            click.echo(f"Created Gateway {gateway_name}")
        else:
            click.echo(f"Error applying Gateway: {e}")
            raise e

def get_tls_gateway_name(domain: str):
    sanitised = domain.replace('.','-')
    return f'{sanitised}-tls-gateway'

def apply_tls_gateway(service_name: str, domain: str, yaml_file_path: str,
                              gateway_namespace: str):
    gateway_name = get_tls_gateway_name(domain=domain)
    with open(yaml_file_path, 'r') as f:
        yaml_content = f.read()
        yaml_dict = yaml.safe_load(yaml_content)
    click.echo(f"Applying Virtual service {gateway_name} for project {service_name}.")
    yaml_dict['metadata']['name'] = gateway_name
    yaml_dict['metadata']['namespace'] = gateway_namespace
    yaml_dict['spec']['servers'][0]['hosts'] = [domain]
    yaml_dict['spec']['servers'][1]['hosts'] = [domain]
    update_or_apply_istio_gateway(
        gateway_name=gateway_name,
        gateway_namespace=gateway_namespace,
        gateway_yaml_dict=yaml_dict
    )


def update_or_apply_virtual_service(virtual_service_name: str, virtual_service_namespace: str,
        virtual_service_yaml_dict: Dict,

):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        raise ValueError("Could not get the cluster context name")
    k8s_api_client = config.new_client_from_config(context=context_name)
    # Create a Kubernetes API client
    k8s_client = client.CustomObjectsApi(k8s_api_client)
    # define the group version and plural for the virtual service
    group = "networking.istio.io"
    version = "v1"
    plural = "virtualservices"

    # Check if the VirtualService already exists
    try:
        existing_service = k8s_client.get_namespaced_custom_object(group=group, version=version,
                                                                   namespace=virtual_service_namespace, plural=plural,
                                                                   name=virtual_service_name)
        resource_version = existing_service['metadata']['resourceVersion']
        # Add the resource_version to the yaml dict
        virtual_service_yaml_dict['metadata']['resourceVersion'] = resource_version
        click.echo(f'Resource version is {resource_version}')
        k8s_client.patch_namespaced_custom_object(group=group, version=version, namespace=virtual_service_namespace,
                                                  plural=plural, name=virtual_service_name,
                                                  body=virtual_service_yaml_dict)
        click.echo(f"Updated VirtualService {virtual_service_name}.")
    except ApiException as e:
        if e.status == 404:
            k8s_client.create_namespaced_custom_object(group=group, version=version,
                                                       namespace=virtual_service_namespace, plural=plural,
                                                       body=virtual_service_yaml_dict)
            click.echo(f"Created VirtualService {virtual_service_name}")
        else:
            click.echo(f"Error applying VirtualService: {e}")
            raise e

def get_subpath_virtual_service_name(service_name:str):
    return f'{service_name}-subpath-ingress'

def apply_virtual_service_for_routing(service_name: str, yaml_file_path: str, sanitised_project_name: str,
                                      env: Optional[str] = None):
    virtual_service_namespace = env if env else DEFAULT_NAMESPACE
    virtual_service_name = get_subpath_virtual_service_name(service_name)

    with open(yaml_file_path, 'r') as f:
        yaml_content = f.read()
        yaml_dict = yaml.safe_load(yaml_content)

    click.echo(f"Applying Virtual service {service_name} for project {sanitised_project_name}.")
    yaml_dict['metadata']['name'] = virtual_service_name
    yaml_dict['metadata']['namespace'] = virtual_service_namespace
    yaml_dict['spec']['hosts'][0] = get_istio_ingress_gateway_hostname()
    yaml_dict['spec']['http'][0]['match'][0]['uri']['prefix'] = f"/svc/{virtual_service_namespace}/{service_name}/"
    yaml_dict['spec']['http'][0]['route'][0]['destination'][
        'host'] = 'knative-local-gateway.istio-system.svc.cluster.local'
    yaml_dict['spec']['http'][0]['route'][0]['headers']['request']['set'][
        'Host'] = f'{service_name}.{virtual_service_namespace}.svc.cluster.local'
    update_or_apply_virtual_service(virtual_service_name=virtual_service_name,
        virtual_service_namespace=virtual_service_namespace, virtual_service_yaml_dict=yaml_dict)


def apply_tls_virtual_service(service_name: str, domain: str, yaml_file_path: str, gateway_name: str,
                              namespace: str) -> bool:
    """
    Apply TLS VirtualService for HTTP to HTTPS redirection
    Args:
        service_name: Name of the service
        domain: Domain name for the service
        yaml_file_path: Path to the TLS virtual service YAML template
        namespace: Kubernetes namespace
        context_name: Kubernetes context name
    Returns:
        bool: True if successful, False otherwise
    """

    virtual_service_name = f'{domain}-tls-ingress'

    # Load and modify YAML
    with open(yaml_file_path, 'r') as f:
        yaml_dict = yaml.safe_load(f.read())

    # Update YAML with service specific values
    yaml_dict['metadata']['name'] = virtual_service_name
    yaml_dict['metadata']['namespace'] = namespace
    yaml_dict['spec']['hosts'][0] = domain
    yaml_dict['spec']['gateways'][0] = f'{namespace}/{gateway_name}'
    yaml_dict['spec']['http'][0]['route'][0]['destination']['host'] = f'{service_name}.{namespace}.svc.cluster.local'
    yaml_dict['spec']['http'][0]['match'][0]['authority']['prefix'] = domain
    yaml_dict['spec']['http'][0]['rewrite']['authority'] = f'{service_name}.{namespace}.svc.cluster.local'

    update_or_apply_virtual_service(virtual_service_name=virtual_service_name, virtual_service_namespace=namespace,
        virtual_service_yaml_dict=yaml_dict)


def apply_knative_service(service_name: str, yaml_file_path: str, sanitised_project_name: str,
                                      image_tag: str, workdir: str, command: str, gpus: int, gpu_type: str, port: int,
                                      domain: Optional[str], env: Optional[str] = None, cpu: float = 100,
                                      memory: int = 200, min_scale: int = 0, max_scale: int = 3, 
                                      context_name: Optional[str] = None, secrets: List[str] = [], enable_efs: bool = False, concurrency: int = 100, readiness: Optional[Dict] = None):

    # Load kube config
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Read the YAML file
    with open(yaml_file_path, 'r') as f:
        yaml_content = f.read()

    click.echo(f"Applying Knative service {service_name} for project {sanitised_project_name} and {gpus} GPUs.")
    yaml_content = yaml_content.replace('${SERVICE_NAME}', service_name)
    yaml_content = yaml_content.replace('${GPUS}', str(gpus))

    # Load the YAML content    
    yaml_dict = yaml.safe_load(yaml_content)

    if enable_efs:
        for volume in yaml_dict['spec']['template']['spec']['volumes']:
            if volume['name'] == 'efs-storage':
                volume['persistentVolumeClaim']['claimName'] = get_efs_claim_name(env_name=env)
    else:
        # Remove the EFS volume
        yaml_dict['spec']['template']['spec']['volumes'] = [volume for volume in yaml_dict['spec']['template']['spec']['volumes'] if volume['name'] != 'efs-storage']
        # Remove the volume mount
        yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'] = [volume for volume in yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'] if volume['name'] != 'efs-storage']
    
    if secrets:
        if enable_efs:
            # podman is ran as a process inside the container so env needs to be read from files and exposed as env vars to podman
            yaml_dict['spec']['template']['spec']['volumes'].append({
                'name': 'secrets',
                'projected': {
                    'sources': [{
                        'secret': {
                            'name': secret_name
                        }
                    } for secret_name in secrets]
                }
            })

            yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'].append({
                'name': 'secrets',
                'mountPath': '/mnt/secrets',
                'readOnly': True
            })
        else:
            yaml_dict['spec']['template']['spec']['containers'][0]['envFrom'] = [{
                'secretRef': {
                    'name': secret_name
                }
            
            } for secret_name in secrets]

    if enable_efs:
        if gpus > 0:
            yaml_dict['spec']['template']['spec']['containers'][0]['image'] = "tensorfuse/podman-nvidia:v1"
            config_nvidia_ctk_commands = """sudo nvidia-ctk cdi generate --output=/etc/cdi/nvidia.yaml
                nvidia-ctk cdi list
                """
            podman_gpu_tags = "--gpus all --env NVIDIA_VISIBLE_DEVICES=all --env NVIDIA_DRIVER_CAPABILITIES=compute,utility"
        else:
            config_nvidia_ctk_commands = ""
            yaml_dict['spec']['template']['spec']['containers'][0]['image'] = "quay.io/podman/stable"
            podman_gpu_tags = ""

        if workdir:
            final_command = f"cd {workdir} && {command}"
        else:
            final_command = command

        if secrets:
            secrets_to_env_vars_command = """\
                folder_path="/mnt/secrets"

                # Initialize an empty string to hold the environment variables
                env_vars=""

                # Loop through each file in the folder
                for file in "$folder_path"/*; do
                    if [[ -f $file ]]; then
                        # Get the filename without the path
                        filename=$(basename "$file")
                        
                        # Get the contents of the file
                        contents=$(<"$file")
                        
                        # Escape the contents to handle any special characters or spaces
                        escaped_contents=$(printf '%q' "$contents")

                        # Append to the env_vars string in the format --env filename=contents
                        env_vars="$env_vars --env $filename=$escaped_contents"
                    fi
                done"""
        else:
            secrets_to_env_vars_command = "env_vars=''"

        yaml_dict['spec']['template']['spec']['containers'][0]['command'] = ["/bin/sh", "-c",
                                                                            f"""{config_nvidia_ctk_commands}
        {secrets_to_env_vars_command}
                sed -i 's|mount_program = "/usr/bin/fuse-overlayfs"|mount_program = ""|' /etc/containers/storage.conf
                sudo podman run --name mycontainer $env_vars {podman_gpu_tags} --network=host \
                --rootfs /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs:O sh -c "{final_command}" """]
    else:
        # if efs is disabled then we dont need podman.
        del yaml_dict['spec']['template']['spec']['containers'][0]['command']
        yaml_dict['spec']['template']['spec']['containers'][0]['image'] = image_tag
        del yaml_dict['spec']['template']['spec']['containers'][0]['securityContext']
        if gpus > 0:
            yaml_dict['spec']['template']['spec']['runtimeClassName'] = get_nydus_runtime_class_name()
            if 'nodeSelector' in yaml_dict['spec']['template']['spec']:
                yaml_dict['spec']['template']['spec']['nodeSelector']['karpenter.sh/nodepool'] = get_gpu_nodepool()
            else:
                yaml_dict['spec']['template']['spec']['nodeSelector'] = {'karpenter.sh/nodepool': get_gpu_nodepool()}


    if gpus > 0:
        # by default gpu pods usually use all of available memory and cpu
        # so I am here defining some limits above which the cpu and memory parameters will be used
        if cpu > 1000:
            yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
                'memory'] = f'{str(int(memory))}M'
        if memory > 2000:
            yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
                'memory'] = f'{str(int(memory))}M'
        yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests']['cpu'] = f'{str(int(cpu))}m'

        if 'nodeSelector' in yaml_dict['spec']['template']['spec']:
            yaml_dict['spec']['template']['spec']['nodeSelector']['karpenter.k8s.aws/instance-family'] = get_instance_family_from_gpu_type(gpu_type)
        else:
            yaml_dict['spec']['template']['spec']['nodeSelector'] = {'karpenter.k8s.aws/instance-family': get_instance_family_from_gpu_type(gpu_type)}
            
        
        yaml_dict['spec']['template']['metadata']['annotations'][
            'autoscaling.knative.dev/scale-to-zero-pod-retention-period'] = '10m'
        yaml_dict['spec']['template']['metadata']['annotations']['autoscaling.knative.dev/scale-down-delay'] = '10m'
        yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'].append(
            {'name': 'dshm', 'mountPath': '/dev/shm'})
        yaml_dict['spec']['template']['spec']['volumes'].append(
            {'name': 'dshm', 'emptyDir': {'medium': 'Memory', 'sizeLimit': '10Gi'}})
    else:
        yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
            'memory'] = f'{str(int(memory))}M'
        yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests']['cpu'] = f'{str(int(cpu))}m'
        apply_anti_affinity_for_gpus(yaml_dict)
    if gpus > 2:
        yaml_dict['spec']['template']['metadata']['annotations'][
            'autoscaling.knative.dev/scale-to-zero-pod-retention-period'] = '20m'
        yaml_dict['spec']['template']['metadata']['annotations']['autoscaling.knative.dev/scale-down-delay'] = '20m'
    # apply min scale and max scale arguements
    yaml_dict['spec']['template']['metadata']['annotations']['autoscaling.knative.dev/min-scale'] = str(min_scale)
    yaml_dict['spec']['template']['metadata']['annotations']['autoscaling.knative.dev/max-scale'] = str(max_scale)
    yaml_dict['spec']['template']['metadata']['annotations']['autoscaling.knative.dev/target'] = str(concurrency)

    yaml_dict['spec']['template']['metadata']['annotations']['image_tag'] = image_tag
    yaml_dict['spec']['template']['metadata']['annotations']['deploy_time'] = datetime.now().isoformat()

    if readiness:
        yaml_dict['spec']['template']['spec']['containers'][0]['readinessProbe'] = readiness
    
    # set container port
    yaml_dict['spec']['template']['spec']['containers'][0]['ports'] = [{'containerPort': port}]
    if readiness is None:
        if 'httpGet' in yaml_dict['spec']['template']['spec']['containers'][0]['readinessProbe']:
            yaml_dict['spec']['template']['spec']['containers'][0]['readinessProbe']['httpGet']['port'] = port

    # enable http redirects if domain is available
    # if domain:
    #     if 'annotations' not in yaml_dict['metadata']:
    #         yaml_dict['metadata']['annotations'] = {}
    #     yaml_dict['metadata']['annotations']['networking.knative.dev/http-protocol'] = 'redirected'

    # Create a Kubernetes API client
    k8s_client = client.CustomObjectsApi(k8s_api_client)

    # Apply the configuration
    group = "serving.knative.dev"
    version = "v1"
    namespace = env if env else DEFAULT_NAMESPACE
    plural = "services"

    # check if the custom object exists if yes then update else create
    try:
        existing_service = k8s_client.get_namespaced_custom_object(group, version, namespace, plural, service_name)
        resource_version = existing_service['metadata']['resourceVersion']
        # add the resource_version to the yaml dict
        yaml_dict['metadata']['resourceVersion'] = resource_version
        click.echo(f'Resource version is {resource_version}')
        # Remove immutable fields
        if 'metadata' in yaml_dict:
            if 'annotations' in yaml_dict['metadata']:
                if 'serving.knative.dev/creator' in yaml_dict['metadata']['annotations']:
                    del yaml_dict['metadata']['annotations']['serving.knative.dev/creator']
            if 'annotations' in yaml_dict['metadata']:
                if 'serving.knative.dev/lastModifier' in yaml_dict['metadata']['annotations']:
                    del yaml_dict['metadata']['annotations']['serving.knative.dev/lastModifier']
        k8s_client.patch_namespaced_custom_object(group, version, namespace, plural, service_name, yaml_dict)
        click.echo(f"Updated Knative service {service_name}.")
    except client.exceptions.ApiException as e:
        if e.status == 404:
            k8s_client.create_namespaced_custom_object(group, version, namespace, plural, yaml_dict)
            click.echo(f"Created Knative service {service_name}.")
        else:
            click.echo(f"Error applying Knative service: {e}")
            raise e


def enable_knative_selectors_pv_pvc_capabilities(namespace=KNATIVE_SERVING_NAMESPACE,
                                                 context_name: Optional[str] = None):
    """
    Enable the nodeSelector feature in the config-features ConfigMap.

    Args:
        namespace (str): The namespace where the ConfigMap is located. Defaults to 'knative-serving'.
    """
    # Load the kubeconfig
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create an instance of the API class
    v1 = client.CoreV1Api(k8s_api_client)

    try:
        # Get the existing config-features ConfigMap
        config_map = v1.read_namespaced_config_map(name=CONFIG_FEATURES, namespace=namespace)

        # Update the ConfigMap data
        if config_map.data is None:
            config_map.data = {}
        config_map.data['kubernetes.podspec-nodeselector'] = 'enabled'
        config_map.data['kubernetes.podspec-affinity'] = 'enabled'
        config_map.data["kubernetes.podspec-persistent-volume-claim"] = "enabled"
        config_map.data["kubernetes.podspec-persistent-volume-write"] = "enabled"
        config_map.data["kubernetes.containerspec-addcapabilities"] = "enabled"
        config_map.data["kubernetes.podspec-security-context"] = "enabled"
        config_map.data["kubernetes.podspec-runtimeclassname"] = "enabled"

        # Update the ConfigMap
        v1.patch_namespaced_config_map(name=CONFIG_FEATURES, namespace=namespace, body=config_map)
        print(
            f"Successfully enabled node selector, affinity, pv-claim, pv-write and add-capabilities features in {CONFIG_FEATURES} ConfigMap.")
    except client.exceptions.ApiException as e:
        print(f"Exception when updating ConfigMap: {e}")
        raise

def add_runtimeclass_knative_features(namespace=KNATIVE_SERVING_NAMESPACE,
                                                 context_name: Optional[str] = None):
    # Load the kubeconfig
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create an instance of the API class
    v1 = client.CoreV1Api(k8s_api_client)

    try:
        # Get the existing config-features ConfigMap
        config_map = v1.read_namespaced_config_map(name=CONFIG_FEATURES, namespace=namespace)

        # Update the ConfigMap data
        if config_map.data is None:
            config_map.data = {}
        config_map.data["kubernetes.podspec-runtimeclassname"] = "enabled"

        # Update the ConfigMap
        v1.patch_namespaced_config_map(name=CONFIG_FEATURES, namespace=namespace, body=config_map)
        print(
            f"Successfully enabled pod runtime feature in {CONFIG_FEATURES} ConfigMap.")
    except client.exceptions.ApiException as e:
        print(f"Exception when updating ConfigMap: {e}")
        raise







def list_deployed_services(env_name: Optional[str] = None, all: bool = False, context_name: Optional[str] = None):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    api_client = config.new_client_from_config(context=context_name)
    api = client.CustomObjectsApi(api_client)
    if all:
        ksvc_list = api.list_cluster_custom_object(group="serving.knative.dev", version="v1", plural="services", )
    else:
        namespace = env_name if env_name else DEFAULT_NAMESPACE
        ksvc_list = api.list_namespaced_custom_object(group="serving.knative.dev", version="v1", plural="services",
                                                      namespace=namespace, )

    return ksvc_list


def get_knative_service(service_name: str, namespace: str = "default", context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    api = client.CustomObjectsApi(k8s_api_client)
    try:
        ksvc = api.get_namespaced_custom_object(group="serving.knative.dev", version="v1", plural="services",
                                                name=service_name, namespace=namespace, )
        return ksvc
    except ApiException as e:
        if e.status == 404:
            return None
        else:
            raise e


def delete_knative_service_by_name(service_name: str, namespace: str = "default", context_name: Optional[str] = None,
                           wait: bool = False, timeout: int = 600) -> bool:
    """
    Delete a Knative service and optionally wait for completion.

    Args:
        service_name: Name of the Knative service
        namespace: Kubernetes namespace
        context_name: Kubernetes context name
        wait: Whether to wait for deletion to complete
        timeout: Maximum time to wait in seconds

    Returns:
        bool: True if deletion successful, False if service not found
    """
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return False

    k8s_api_client = config.new_client_from_config(context=context_name)
    api = client.CustomObjectsApi(k8s_api_client)

    try:
        # Delete the Knative service
        api.delete_namespaced_custom_object(
            group="serving.knative.dev",
            version="v1",
            plural="services",
            name=service_name,
            namespace=namespace,
            body=client.V1DeleteOptions(
                propagation_policy="Background",
                grace_period_seconds=0
            )
        )

        if wait:
            start_time = time.time()
            while time.time() - start_time < timeout:
                try:
                    api.get_namespaced_custom_object(
                        group="serving.knative.dev",
                        version="v1",
                        plural="services",
                        name=service_name,
                        namespace=namespace
                    )
                    time.sleep(2)
                except ApiException as e:
                    if e.status == 404:
                        return True
                    raise e
            raise TimeoutError(f"Timeout waiting for service {service_name} deletion")

        return True

    except ApiException as e:
        if e.status == 404:
            return False
        raise e


def get_ready_condition(service):
    ready_condition = [condition for condition in service['status']['conditions'] if condition['type'] == 'Ready']
    if ready_condition:
        return ready_condition[0]
    return None


def get_latest_running_revision(service_name: str, namespace: str = "default"):
    service = get_knative_service(service_name, namespace)
    if service is None:
        return None
    latest_revision_name = service['status'].get('latestReadyRevisionName', None)
    return latest_revision_name


def get_pods_for_service(service_name: str, namespace: str = "default"):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    api = client.CoreV1Api(k8s_api_client)
    latest_revision = get_latest_running_revision(service_name, namespace)
    if latest_revision is None:
        return None
    pods = api.list_namespaced_pod(namespace=namespace,
                                   label_selector=f"serving.knative.dev/service={service_name},serving.knative.dev/revision={latest_revision}")
    return pods


def delete_knative_services(context_name: Optional[str] = None):
    # kubectl delete ksvc --all -n <your-namespace>
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    command = ["kubectl", "--context", f"{context_name}", "delete", "ksvc", "--all", "-n", "default"]
    subprocess.run(command, check=True)  # TODO maybe wait for pods to scale down before returning


def cleanup_knative_resources(context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    try:
        # kubectl delete gateway --all -n istio-system
        command = ["kubectl", "--context", f"{context_name}", "delete", "gateway", "--all", "-n", "istio-system"]
        subprocess.run(command, check=True)
        #  kubectl delete gateway --all -n knative-serving
        command = ["kubectl", "--context", f"{context_name}", "delete", "gateway", "--all", "-n", "knative-serving"]
        subprocess.run(command, check=True)
    except Exception as e:
        click.echo(f"Error while cleaning up Istio gateways: {e}")


def list_ksvc_in_namespace(namespace: str, context_name: Optional[str] = None):
    # Load the Kubernetes configuration
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create an instance of the CustomObjectsApi
    api_instance = client.CustomObjectsApi(k8s_api_client)

    # Define the group, version, and plural for the Knative service
    group = 'serving.knative.dev'
    version = 'v1'
    plural = 'services'

    try:
        # List all Knative services in the specified namespace
        services = api_instance.list_namespaced_custom_object(group=group, version=version, namespace=namespace,
                                                              plural=plural)
        return services
    except ApiException as e:
        print(f"Failed to list Knative services in namespace {namespace}: {e}")
        raise e


def delete_ksvc_from_namespace(service_name: str, namespace: str, context_name: Optional[str] = None):
    # Load the Kubernetes configuration
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create an instance of the CustomObjectsApi
    api_instance = client.CustomObjectsApi(k8s_api_client)

    # Define the group, version, and plural for the Knative service
    group = 'serving.knative.dev'
    version = 'v1'
    plural = 'services'

    try:
        # Delete the Knative service
        api_instance.delete_namespaced_custom_object(group=group, version=version, namespace=namespace, plural=plural,
                                                     name=service_name, body=client.V1DeleteOptions())
        print(f"Knative service {service_name} deletion initiated.")
        # Wait for the service to be deleted
        while True:
            try:
                api_instance.get_namespaced_custom_object(group, version, namespace, plural, service_name)
                time.sleep(1)  # Wait for 1 second before checking again
            except ApiException as e:
                if e.status == 404:
                    print(f"Knative service {service_name} deleted successfully.")
                    break
                else:
                    print(f"Error while waiting for deletion of Knative service {service_name}: {e}")
                    raise e
        print(f"Knative service {service_name} deleted successfully.")
    except ApiException as e:
        print(f"Failed to delete Knative service {service_name}: {e}")


def delete_all_ksvc_from_namespace(namespace: str):
    if namespace in TENSORFUSE_NAMESPACES:
        click.echo(f"Namespace {namespace} is a system namespace. Skipping deletion of Knative services.")
        return False
    services = list_ksvc_in_namespace(namespace)
    for service in services['items']:
        service_name = service['metadata']['name']
        delete_ksvc_from_namespace(service_name, namespace)


def get_istio_ingress_gateway_hostname():
    # Load the Kubernetes configuration

    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    api_client = config.new_client_from_config(context=context_name)

    # Create an instance of the API class
    api_instance = client.CoreV1Api(api_client)

    try:
        # Get the list of services in the istio-system namespace
        services = api_instance.list_namespaced_service(namespace='istio-system')

        # Find the Istio ingress gateway service
        ingress_gateway = next((svc for svc in services.items if svc.metadata.name.startswith('istio-ingressgateway')),
                               None)

        if ingress_gateway and ingress_gateway.status.load_balancer.ingress:
            # Return the hostname or IP of the load balancer
            return ingress_gateway.status.load_balancer.ingress[0].hostname or \
                ingress_gateway.status.load_balancer.ingress[0].ip
        else:
            print("Istio ingress gateway not found or does not have a load balancer.")
            return None
    except client.exceptions.ApiException as e:
        print(f"Exception when calling CoreV1Api->list_namespaced_service: {e}")
        return None

def delete_virtual_service(virtual_service_name: str, namespace: str):
    """
    Deletes the specified VirtualService.
    """
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    api = client.CustomObjectsApi(k8s_api_client)

    group = "networking.istio.io"
    version = "v1"
    plural = "virtualservices"

    try:
        api.delete_namespaced_custom_object(
            group=group,
            version=version,
            namespace=namespace,
            plural=plural,
            name=virtual_service_name,
            body=client.V1DeleteOptions()
        )
        click.echo(f"VirtualService {virtual_service_name} deletion initiated.")
    except ApiException as e:
        if e.status == 404:
            click.echo(f"VirtualService {virtual_service_name} not found.")
        else:
            click.echo(f"Failed to delete VirtualService {virtual_service_name}: {e}")


def check_existing_virtual_service(virtual_service_name: str, namespace: str) -> bool:
    """
    Checks if a VirtualService with the given name exists in the specified namespace.
    Returns True if it exists, otherwise False.
    """
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return False
    k8s_api_client = config.new_client_from_config(context=context_name)
    api = client.CustomObjectsApi(k8s_api_client)

    group = "networking.istio.io"
    version = "v1"
    plural = "virtualservices"

    try:
        # Attempt to get the specified VirtualService
        api.get_namespaced_custom_object(
            group=group,
            version=version,
            namespace=namespace,
            plural=plural,
            name=virtual_service_name
        )
        print(f"VirtualService {virtual_service_name} exists in namespace {namespace}.")
        return True
    except ApiException as e:
        if e.status == 404:
            print(f"VirtualService {virtual_service_name} does not exist in namespace {namespace}.")
            return False
        else:
            print(f"Error checking VirtualService {virtual_service_name}: {e}")
            return False


import click

from tensorkube.constants import get_cluster_name, get_job_sidecar_iam_role_name
from tensorkube.services.aws_service import get_aws_account_id
from tensorkube.services.dynamodb_service import get_dynamodb_access_policy_name
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url
from tensorkube.services.iam_service import attach_role_policy, \
    create_or_update_iam_role_with_service_account_cluster_access
from tensorkube.services.k8s_service import patch_service_account
from tensorkube.services.s3_access_service import get_s3_access_policy_name, update_s3_policy
from tensorkube.services.sqs_service import get_sqs_access_policy_name


def apply():
    try:
        s3_access_policy_name = get_s3_access_policy_name()
        update_s3_policy(policy_name=s3_access_policy_name)
        dynamodb_access_policy_name = get_dynamodb_access_policy_name()
        sqs_access_policy_name = get_sqs_access_policy_name()
        role_name = get_job_sidecar_iam_role_name()
        job_queue_sidecar_sa_name = "job-queue-sidecar-keda-sa"
        role = create_or_update_iam_role_with_service_account_cluster_access(get_aws_account_id(),
                                                                             get_cluster_oidc_issuer_url(get_cluster_name()),
                                                                             role_name, job_queue_sidecar_sa_name,
                                                                             'keda')
        role_arn = role['Role']['Arn']
        attach_role_policy(account_no=get_aws_account_id(), policy_name=s3_access_policy_name, role_name=role_name)
        attach_role_policy(account_no=get_aws_account_id(), policy_name=dynamodb_access_policy_name, role_name=role_name)
        attach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_access_policy_name, role_name=role_name)
        patch_service_account(role_arn, job_queue_sidecar_sa_name, "keda")
        click.echo("Successfully migrated to ISRA for fine-tuning s3 access")


    except Exception as e:
        raise e



import os
import time
import uuid
from typing import Optional, Dict, List

import click
import yaml
from kubernetes import client
from pkg_resources import resource_filename
from tensorkube.services.k8s_service import find_and_delete_old_job, check_pod_status, start_streaming_pod,  \
    get_pod_name_corresponing_to_job, get_pod_status_from_job, list_secrets, get_image_tags_to_retain , apply_image_cleanup_job, start_streaming_service
from tensorkube.helpers import  sanitise_name, \
    extract_command_from_dockerfile, extract_workdir_from_dockerfile

from tensorkube.constants import DEFAULT_NAMESPACE, BUILD_TOOL, PodStatus, CliColors
from tensorkube.services.aws_service import get_credentials, get_bucket_name, are_credentials_valid
from tensorkube.services.environment_service import check_environment_exists
from tensorkube.services.s3_service import  upload_files_in_parallel, empty_s3_folder

from tensorkube.services.build import apply_k8s_buildkit_config

from tensorkube.services.knative_service import apply_knative_service, apply_virtual_service_for_routing, \
    apply_tls_virtual_service, apply_tls_gateway, get_tls_gateway_name, check_existing_virtual_service, \
    get_subpath_virtual_service_name, delete_virtual_service


DEFAULT_GPUS = 0
DEFAULT_GPU_TYPE = None
DEFAULT_CPU = 100
DEFAULT_MEMORY = 200
DEFAULT_MIN_SCALE = 0
DEFAULT_MAX_SCALE = 3
DEFAULT_GITHUB_ACTIONS = False
DEFAULT_ENV = None
DEFAULT_SECRET = None


class Config(object):
    def __init__(self, config_file_path):
        self.__config = {}
        if config_file_path != None:
            file_path = os.path.abspath(config_file_path)  # Convert to absolute path
            if not os.path.isfile(file_path):
                raise Exception(f"Config.yaml file not found in path: {file_path}")
            with open(file_path, 'r') as config_file:
                self.__config = yaml.safe_load(config_file)

    def get(self, key):
        return self.__config.get(key, None)

    def set_if_not_exist(self, key, value):
        if key in self.__config:
            return self
        if isinstance(value, tuple):
            self.__config[key] = list(value)
        else:
            self.__config[key] = value
        return self

    def __str__(self) -> str:
        return str(self.__config)


def upload_build_context_to_s3(bucket_name, sanitised_project_name, cwd):
    # TODO: figure out how to upload only the updated files to the s3 bucket
    click.echo("Deleting any outdated files in the S3 bucket...")
    empty_s3_folder(bucket_name=bucket_name, folder_name="build/" + sanitised_project_name + "/")

    click.echo("Uploading the current directory to the S3 bucket...")
    upload_files_in_parallel(bucket_name=bucket_name, folder_path=cwd, s3_path="build/" + sanitised_project_name + "/")


def create_deployment_details(secret: list, env: str):
    secrets = [s for s in secret]
    if env and not check_environment_exists(env_name=env):
        click.echo("Environment does not exist. Please create the environment first.")
        return None, None, None, None, None
    env_namespace = DEFAULT_NAMESPACE if not env else env
    if secrets:
        secrets_list = list_secrets(env_namespace)
        secrets_names_list = [secret.metadata.name for secret in secrets_list]
        for secret in secrets:
            if secret not in secrets_names_list:
                click.echo(
                    click.style("Secret ", fg=CliColors.ERROR.value) + click.style(secret, fg=CliColors.ERROR.value,
                                                                                   bold=True,
                                                                                   italic=True) + click.style(
                        " not found in the environment.", fg=CliColors.ERROR.value))
                return None, None, None, None, None

    image_tag = uuid.uuid4().hex
    cwd = os.getcwd()
    non_sanitised_name = os.path.basename(cwd)
    sanitised_project_name = sanitise_name(non_sanitised_name)

    return secrets, env_namespace, image_tag, cwd, sanitised_project_name


def build_app(env, sanitised_project_name, image_tag, upload_to_nfs: bool = False, image_url: Optional[str] = None, convert_to_nydus_image: bool= False,
              work_dir: str = os.getcwd(), secrets: List[str] = []):
    is_dockerfile_present = False
    for root, dirs, files in os.walk(work_dir):
        for file in files:
            local_file = os.path.join(root, file)
            if local_file == work_dir + "/Dockerfile":
                is_dockerfile_present = True

    if not is_dockerfile_present:
        click.echo("No Dockerfile found in current directory.")
        return

    bucket_name = get_bucket_name(env_name=env)

    build_job_name = f'{BUILD_TOOL}-{sanitised_project_name}'
    old_cleanup_job = 'cleanup-{}'.format(sanitised_project_name)

    env_namespace = DEFAULT_NAMESPACE if not env else env
    old_job_deleted = find_and_delete_old_job(job_name=build_job_name, namespace=env_namespace)
    old_cleanup_job_del = find_and_delete_old_job(job_name=old_cleanup_job, namespace=env_namespace)
    if not old_job_deleted:
        click.echo("Another deployment is already in progress. Please wait for the build to complete.")
        return

    # TODO!: add logic to update the aws-secret only if IAM Identity Center User
    credentials = get_credentials()
    if are_credentials_valid(credentials):
        click.echo("Credentials are up to date")
    else:
        click.echo("The AWS credentials have expired. Please update the credentials.")
        return
    upload_build_context_to_s3(bucket_name=bucket_name, sanitised_project_name=sanitised_project_name, cwd=work_dir)
    click.echo("Building the Docker image...")
    
    apply_k8s_buildkit_config(sanitised_project_name=sanitised_project_name, image_tag=image_tag, env_name=env, image_url=image_url, upload_to_nfs=upload_to_nfs,convert_to_nydus_image=convert_to_nydus_image, secrets=secrets)
    build_job_pod_name = get_pod_name_corresponing_to_job(job_name=build_job_name, namespace=env_namespace)
    if build_job_pod_name is None:
        click.echo("Build Pod not found")
        return PodStatus.FAILED.value
    # TODO: stream multiple lines instead of one by one
    start_streaming_pod(pod_name=build_job_pod_name, namespace=env_namespace, status=PodStatus.SUCCEEDED)

    transition_time = time.time()
    # wait for the pod to transition
    while True:
        try:
            pod_status = check_pod_status(pod_name=build_job_pod_name, namespace=env_namespace)
        except client.ApiException as e:
            if e.status == 404:
                print('Pod not found.')
                pod_status = get_pod_status_from_job(job_name=build_job_name, namespace=env_namespace)
            else:
                pod_status = PodStatus.FAILED.value
        print('Waiting for pod to transition')
        if pod_status in ['Succeeded', 'Completed', 'Failed']:
            break
        if time.time() - transition_time > 60:  # 60 seconds have passed
            print("Timeout: Pod did not reach the desired state within 1 minute.")
            break
        time.sleep(5)
    return pod_status

def deploy_knative_service(env: str, gpus: int, gpu_type: str, cpu: float, memory: float, min_scale: int, max_scale: int, concurrency: int,
                           pod_status: str, cwd: str, sanitised_project_name: str, image_url: str, secrets: list, port: int, domain: Optional[str],
                           enable_efs: bool = False, github_actions: bool = DEFAULT_GITHUB_ACTIONS, readiness: Optional[Dict] = None):
    env_namespace = DEFAULT_NAMESPACE if not env else env

    if pod_status == 'Succeeded' or pod_status == 'Completed':
        click.echo("Successfully built and pushed the Docker image.")
        yaml_file_path = resource_filename('tensorkube', 'configurations/build_configs/knative_base_config.yaml')
        dockerfile_path = cwd + "/Dockerfile"
        workdir = extract_workdir_from_dockerfile(dockerfile_path)
        command = extract_command_from_dockerfile(dockerfile_path)
        service_name = f"{sanitised_project_name}-gpus-{gpus}-{str(gpu_type).lower()}"
        apply_knative_service(service_name=service_name, yaml_file_path=yaml_file_path,
                                            sanitised_project_name=sanitised_project_name, image_tag=image_url,
                                            workdir=workdir, command=command, gpus=gpus, gpu_type=gpu_type, cpu=cpu,
                                            memory=memory, min_scale=min_scale, max_scale=max_scale, env=env,
                                            secrets=secrets, enable_efs=enable_efs, domain=domain, concurrency=concurrency, readiness = readiness, port = port)
        if domain is None:
            virtual_service_yaml_file_path = resource_filename('tensorkube',
                                                               'configurations/build_configs/virtual_service.yaml')
            apply_virtual_service_for_routing(service_name=service_name, yaml_file_path=virtual_service_yaml_file_path,
                                              env=env, sanitised_project_name=sanitised_project_name)
        else:
            # check if the subpath virtual service exists if yes clean it up
            click.echo(click.style(
                "Checking for existing virtual service for subpath routing", fg="green"
            ))
            subpath_virtual_service_name = get_subpath_virtual_service_name(
                                                      service_name=service_name)
            if check_existing_virtual_service(namespace=env_namespace,
                                              virtual_service_name=subpath_virtual_service_name):
                delete_virtual_service(virtual_service_name=subpath_virtual_service_name, namespace=env_namespace)
            virtual_service_yaml_file_path = resource_filename('tensorkube',
                                                               'configurations/build_configs/istio_virtualservice_domain.yaml')
            gateway_yaml_file_path = resource_filename('tensorkube',
                                                       'configurations/build_configs/istio_gateway_subdomain_base.yaml')
            # Apply the Istio Gateway
            apply_tls_gateway(service_name=service_name, domain=domain, yaml_file_path=gateway_yaml_file_path,
                              gateway_namespace=env_namespace)
            apply_tls_virtual_service(service_name=service_name, domain=domain,
                                      yaml_file_path=virtual_service_yaml_file_path,
                                      gateway_name=get_tls_gateway_name(domain), namespace=env_namespace)
        images_to_retain = get_image_tags_to_retain(sanitised_project_name=sanitised_project_name,
                                                    service_name=service_name, namespace=env_namespace)
        if enable_efs:
            apply_image_cleanup_job(sanitised_project_name=sanitised_project_name, image_tags=images_to_retain, env=env)
        start_streaming_service(service_name=service_name, namespace=env_namespace)
    else:
        if github_actions:
            raise Exception("Failed to build the Docker image. Please check the logs for more details. Pod status: {}".format(
                pod_status))
        else:
            click.echo("Failed to build the Docker image. Please check the logs for more details. Pod status: {}".format(
                pod_status))



import json
import os
import platform
import subprocess
from string import Template
from typing import List
import yaml
import click
import semver
from pkg_resources import resource_filename

from tensorkube.constants import CliColors, LOCKED_EKSCTL_VERSION
from tensorkube.constants import get_cluster_name
from tensorkube.services.aws_service import get_aws_account_id, get_karpenter_namespace
from tensorkube.services.error import CLIVersionError


# create base cluster using eksctl
def create_base_tensorkube_cluster_eksctl(cluster_name, vpc_public_subnets: List[str] = [], vpc_private_subnets: List[str]=[]):
    yaml_file_path = resource_filename('tensorkube', 'configurations/base_cluster.yaml')
    # variables
    variables = {"CLUSTER_NAME": cluster_name, "AWS_DEFAULT_REGION": "us-east-1", "K8S_VERSION": "1.29",
                 "AWS_ACCOUNT_ID": get_aws_account_id(), "KARPENTER_NAMESPACE": get_karpenter_namespace(),
                 "AWS_PARTITION": "aws", }

    with open(yaml_file_path) as file:
        template = file.read()
    yaml_content = Template(template).substitute(variables)

    yaml_data = yaml.safe_load(yaml_content)
    
    # Dynamically construct the VPC section
    vpc_config = {"subnets": {"private": {}, "public": {}}}
    
    # Add private subnets
    for i, subnet_id in enumerate(vpc_private_subnets, start=1):
        vpc_config["subnets"]["private"][f"subnetprivate{i}"] = {"id": subnet_id}

    # Add public subnets
    for i, subnet_id in enumerate(vpc_public_subnets, start=1):
        vpc_config["subnets"]["public"][f"subnetpublic{i}"] = {"id": subnet_id}

    if len(vpc_public_subnets) > 1 and len(vpc_private_subnets) > 1:
        yaml_data["vpc"] = vpc_config

    temp_yaml_file_path = "/tmp/temp_cluster.yaml"
    with open(temp_yaml_file_path, 'w') as file:
        yaml.dump(yaml_data, file, default_flow_style=False)

    # Check if the cluster already exists
    try:
        subprocess.run(["eksctl", "get", "cluster", cluster_name], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        print(f'Cluster {cluster_name} already exists.')
        os.remove(temp_yaml_file_path)
        return None
    except subprocess.CalledProcessError:
        pass  # Cluster does not exist, we can create it

        # Run the eksctl create cluster command
    command = ["eksctl", "create", "cluster", "-f", temp_yaml_file_path]
    subprocess.run(command, check=True)

    # Remove the temporary file
    os.remove(temp_yaml_file_path)


def delete_cluster():
    command = ["eksctl", "delete", "cluster", "--name", get_cluster_name()]
    subprocess.run(command)


def check_and_install_eksctl():
    """Check if eksctl is installed and if not isntall it."""
    try:
        version_cmd = subprocess.run(["eksctl", "version", "-o", "json"], check=True, stdout=subprocess.PIPE)
        version_dict = json.loads(version_cmd.stdout.decode("utf-8"))
        version = semver.VersionInfo.parse(version_dict["Version"])
        locked_eksctl_version = semver.VersionInfo.parse(LOCKED_EKSCTL_VERSION)
        # check if the eksctl version is above locked eksctl version
        print("eksctl is already installed.")
        if (version.major, version.minor) < (locked_eksctl_version.major, locked_eksctl_version.minor):
            text = f"eksctl version is {version}. Please upgrade eksctl to version above {LOCKED_EKSCTL_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("Eksctl not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        # check if the operating system is mac and install eksctl
        if platform.system() == "Darwin" or platform.system() == "Linux":
            try:
                loc = f"https://github.com/eksctl-io/eksctl/releases/download/v{LOCKED_EKSCTL_VERSION}/eksctl_Darwin_amd64.tar.gz"
                download_command = ["curl", "--silent", "--location", loc, "-o", "/tmp/eksctl.tar.gz"]
                subprocess.run(" ".join(download_command), shell=True, check=True)
                # Extract the tar.gz file
                extract_command = ["tar", "xzf", "/tmp/eksctl.tar.gz", "-C", "/tmp"]
                subprocess.run(extract_command, check=True)
                # Move the binary to /usr/local/bin
                move_command = ["sudo", "mv", "/tmp/eksctl", "/usr/local/bin"]
                subprocess.run(move_command, check=True)
                subprocess.run(["eksctl", "version"])
            except Exception as e:
                print("Unable to install eksctl. Please install eksctl manually.")
                raise e
        else:
            print("eksctl is not installed. Please install eksctl manually.")
            raise e


def delete_nodegroup(nodegroup_name):
    # Construct the eksctl command to delete the nodegroup
    delete_command = f"eksctl delete nodegroup --name={nodegroup_name} --cluster={get_cluster_name()} --wait"

    # Execute the command
    print(f"Initiating deletion of nodegroup {nodegroup_name}...")
    subprocess.run(delete_command, shell=True, check=True)
    print(f"Nodegroup {nodegroup_name} deletion initiated. Please wait for the process to complete.")



import click

from tensorkube.services.istio import install_istio_on_cluster


def apply():
    try:
        install_istio_on_cluster()
        click.echo("Successfully updated Istio version")

    except Exception as e:
        raise e